@@ -190,7 +190,8 @@ static int convert_rgs(struct gfs2_sbd *sbp)
 		sbp->dinodes_alloced += rgd1->rg_useddi;
 		convert_bitmaps(sbp, rgd);
 		/* Write the updated rgrp to the gfs2 buffer */
-		gfs2_rgrp_out_bh(&rgd->rg, rgd->bits[0].bi_bh);
+		gfs2_rgrp_out(&rgd->rg, rgd->bits[0].bi_bh->b_data);
+		bmodified(rgd->bits[0].bi_bh);
 		rgs++;
 		if (rgs % 100 == 0) {
 			printf(".");
@@ -312,7 +313,8 @@ static void fix_metatree(struct gfs2_sbd *sbp, struct gfs2_inode *ip,
 			bh = bread(sbp, block);
 			if (new)
 				memset(bh->b_data, 0, sbp->bsize);
-			gfs2_meta_header_out_bh(&mh, bh);
+			gfs2_meta_header_out(&mh, bh->b_data);
+			bmodified(bh);
 		}
 
 		hdrsize = blk->height ? sizeof(struct gfs2_meta_header) :
@@ -393,7 +395,7 @@ static void fix_metatree(struct gfs2_sbd *sbp, struct gfs2_inode *ip,
 /*                                                                           */
 /* Returns: 0 on success, -1 on failure                                      */
 /*                                                                           */
-/* Adapted from gfs2_fsck metawalk.c's build_and_check_metalist              */
+/* Adapted from fsck.gfs2 metawalk.c's build_and_check_metalist              */
 /* ------------------------------------------------------------------------- */
 
 static void jdata_mp_gfs1_to_gfs2(struct gfs2_sbd *sbp, int gfs1_h, int gfs2_h,
@@ -469,8 +471,10 @@ static uint64_t fix_jdatatree(struct gfs2_sbd *sbp, struct gfs2_inode *ip,
 			bh = bread(sbp, block);
 			if (new)
 				memset(bh->b_data, 0, sbp->bsize);
-			if (h < (blk->height - 1))
-				gfs2_meta_header_out_bh(&mh, bh);
+			if (h < (blk->height - 1)) {
+				gfs2_meta_header_out(&mh, bh->b_data);
+				bmodified(bh);
+			}
 		}
 
 		if (amount > sbp->bsize - ptramt)
@@ -1070,11 +1074,11 @@ static int inode_renumber(struct gfs2_sbd *sbp, uint64_t root_inode_addr, osi_li
 			bh = bread(sbp, block);
 			if (!gfs2_check_meta(bh, GFS_METATYPE_DI)) {/* if it is an dinode */
 				/* Skip the rindex and jindex inodes for now. */
-				if (block != rindex_addr && block != jindex_addr)
+				if (block != rindex_addr && block != jindex_addr) {
 					error = adjust_inode(sbp, bh);
-					if (error) {
+					if (error)
 						return error;
-					}
+				}
 			} else { /* It's metadata, but not an inode, so fix the bitmap. */
 				int blk, buf_offset;
 				int bitmap_byte; /* byte within the bitmap to fix */
@@ -1299,7 +1303,7 @@ static int fix_one_directory_exhash(struct gfs2_sbd *sbp, struct gfs2_inode *dip
 				 (unsigned long long)leaf_block);
 			break;
 		}
-		gfs2_leaf_in(&leaf, bh_leaf); /* buffer to structure */
+		gfs2_leaf_in(&leaf, bh_leaf->b_data);
 		error = process_dirent_info(dip, sbp, bh_leaf, leaf.lf_entries, dentmod);
 		bmodified(bh_leaf);
 		brelse(bh_leaf);
@@ -1584,7 +1588,7 @@ static int init(struct gfs2_sbd *sbp, struct gfs2_options *opts)
 	bh = bread(sbp, GFS2_SB_ADDR >> sbp->sd_fsb2bb_shift);
 	memcpy(&raw_gfs1_ondisk_sb, (struct gfs_sb *)bh->b_data,
 		   sizeof(struct gfs_sb));
-	gfs2_sb_in(&sbp->sd_sb, bh);
+	gfs2_sb_in(&sbp->sd_sb, bh->b_data);
 
 	jindex_addr = be64_to_cpu(raw_gfs1_ondisk_sb.sb_jindex_di.no_addr);
 	rindex_addr = be64_to_cpu(raw_gfs1_ondisk_sb.sb_rindex_di.no_addr);
@@ -1875,9 +1879,10 @@ static int journ_space_to_rg(struct gfs2_sbd *sdp)
 		convert_bitmaps(sdp, rgd);
 		for (x = 0; x < rgd->ri.ri_length; x++) {
 			if (x)
-				gfs2_meta_header_out_bh(&mh, rgd->bits[x].bi_bh);
+				gfs2_meta_header_out(&mh, rgd->bits[x].bi_bh->b_data);
 			else
-				gfs2_rgrp_out_bh(&rgd->rg, rgd->bits[x].bi_bh);
+				gfs2_rgrp_out(&rgd->rg, rgd->bits[x].bi_bh->b_data);
+			bmodified(rgd->bits[x].bi_bh);
 		}
 	} /* for each journal */
 	return error;
@@ -2123,8 +2128,7 @@ static void copy_quotas(struct gfs2_sbd *sdp)
 
 static int gfs2_query(struct gfs2_options *opts, const char *dev)
 {
-	char response[3] = { 0, 0 };
-	int ret = 0;
+	int res = 0;
 
 	if(opts->yes)
 		return 1;
@@ -2133,27 +2137,25 @@ static int gfs2_query(struct gfs2_options *opts, const char *dev)
 
 	opts->query = TRUE;
 	while (1) {
+		char *line = NULL;
+		size_t len = 0;
+		int ret;
+
 		printf(_("Convert %s from GFS1 to GFS2? (y/n)"), dev);
-		/* Make sure query is printed out */
-		fflush(NULL);
-		fgets(response, 3, stdin);
-		printf("\n");
-		fflush(NULL);
-		response[1] = 0;
-		ret = rpmatch(response);
-
-		if (ret >= 0)
+		fflush(stdout);
+		ret = getline(&line, &len, stdin);
+		res = rpmatch(line);
+		free(line);
+		if (ret <= 0)
+			continue;
+		if (res == 1 || res == 0)
 			break;
-		printf(_("Bad response '%s', please type 'y' or 'n'.\n"), response);
+		/* Unrecognized input; go again. */
 	}
-
 	opts->query = FALSE;
-	return ret;
+	return res;
 }
 
-/* ------------------------------------------------------------------------- */
-/* main - mainline code                                                      */
-/* ------------------------------------------------------------------------- */
 int main(int argc, char **argv)
 {
 	int error;
@@ -2173,7 +2175,7 @@ int main(int argc, char **argv)
 	 */
 	if (!error) {
 		if (sanity_check(&sb2)) {
-			log_crit(_("%s is not a clean gfs filesytem. Please use the"
+			log_crit(_("%s is not a clean gfs filesystem. Please use the"
 				   " fsck.gfs2 utility to correct these errors and"
 				   " try again.\n"), opts.device);
 			exit(0);
@@ -2321,7 +2323,7 @@ int main(int argc, char **argv)
 		{
 			struct gfs2_inode *ip = sb2.master_dir;
 			if (ip->i_bh->b_modified)
-				gfs2_dinode_out(&ip->i_di, ip->i_bh);
+				gfs2_dinode_out(&ip->i_di, ip->i_bh->b_data);
 		}
 		/* Copy old quotas */
 		copy_quotas(&sb2);
@@ -2346,7 +2348,7 @@ int main(int argc, char **argv)
 		/* end because if the tool is interrupted in the middle, we want */
 		/* it to not reject the partially converted fs as already done   */
 		/* when it's run a second time.                                  */
-		bh = bread(&sb2, sb2.sb_addr);
+		bh = bread(&sb2, LGFS2_SB_ADDR(&sb2));
 		sb2.sd_sb.sb_fs_format = GFS2_FORMAT_FS;
 		sb2.sd_sb.sb_multihost_format = GFS2_FORMAT_MULTI;
 		gfs2_sb_out(&sb2.sd_sb, bh->b_data);
