@@ -852,7 +852,7 @@ static void set_rgrp_flags(int rgnum, uint32_t new_flags, int modify, int full)
 	if (sbd.gfs1)
 		gfs_rgrp_in(&rg.rg1, rbh);
 	else
-		gfs2_rgrp_in(&rg.rg2, rbh);
+		gfs2_rgrp_in(&rg.rg2, rbh->b_data);
 	if (modify) {
 		printf("RG #%d (block %llu / 0x%llx) rg_flags changed from 0x%08x to 0x%08x\n",
 		       rgnum, (unsigned long long)rgblk,
@@ -861,7 +861,8 @@ static void set_rgrp_flags(int rgnum, uint32_t new_flags, int modify, int full)
 		if (sbd.gfs1)
 			gfs_rgrp_out(&rg.rg1, rbh);
 		else
-			gfs2_rgrp_out_bh(&rg.rg2, rbh);
+			gfs2_rgrp_out(&rg.rg2, rbh->b_data);
+		bmodified(rbh);
 		brelse(rbh);
 	} else {
 		if (full) {
@@ -973,7 +974,7 @@ static void read_superblock(int fd)
 	sbd.qcsize = GFS2_DEFAULT_QCSIZE;
 	sbd.time = time(NULL);
 	sbd.rgtree.osi_node = NULL;
-	gfs2_sb_in(&sbd.sd_sb, bh); /* parse it out into the sb structure */
+	gfs2_sb_in(&sbd.sd_sb, bh->b_data);
 	/* Check to see if this is really gfs1 */
 	if (sbd1->sb_fs_format == GFS_FORMAT_FS &&
 		sbd1->sb_header.mh_type == GFS_METATYPE_SB &&
@@ -1055,7 +1056,7 @@ static int read_master_dir(void)
 	bh = bread(&sbd, sbd.sd_sb.sb_master_dir.no_addr);
 	if (bh == NULL)
 		return 1;
-	gfs2_dinode_in(&di, bh); /* parse disk inode into structure */
+	gfs2_dinode_in(&di, bh->b_data);
 	do_dinode_extended(&di, bh); /* get extended data, if any */
 	memcpy(&masterdir, &indirect[0], sizeof(struct indirect_info));
 	return 0;
@@ -1103,7 +1104,7 @@ int display(int identify_only, int trunc_zeros, uint64_t flagref,
 	indirect_blocks = 0;
 	lines_per_row[dmode] = 1;
 	if (gfs2_struct_type == GFS2_METATYPE_SB || blk == 0x10 * (4096 / sbd.bsize)) {
-		gfs2_sb_in(&sbd.sd_sb, bh); /* parse it out into the sb structure */
+		gfs2_sb_in(&sbd.sd_sb, bh->b_data);
 		memset(indirect, 0, sizeof(struct iinfo));
 		indirect->ii[0].block = sbd.sd_sb.sb_master_dir.no_addr;
 		indirect->ii[0].is_dir = TRUE;
@@ -1126,7 +1127,7 @@ int display(int identify_only, int trunc_zeros, uint64_t flagref,
 		indirect->ii[0].dirent[1].dirent.de_type = DT_DIR;
 	}
 	else if (gfs2_struct_type == GFS2_METATYPE_DI) {
-		gfs2_dinode_in(&di, bh); /* parse disk inode into structure */
+		gfs2_dinode_in(&di, bh->b_data);
 		do_dinode_extended(&di, bh); /* get extended data, if any */
 	}
 	else if (gfs2_struct_type == GFS2_METATYPE_IN) { /* indirect block list */
@@ -1708,7 +1709,7 @@ static void find_print_block_rg(int bitmap)
 	struct rgrp_tree *rgd;
 
 	rblock = blockstack[blockhist % BLOCK_STACK_SIZE].block;
-	if (rblock == sbd.sb_addr)
+	if (rblock == LGFS2_SB_ADDR(&sbd))
 		printf("0 (the superblock is not in the bitmap)\n");
 	else {
 		rgd = gfs2_blk2rgrpd(&sbd, rblock);
@@ -1762,7 +1763,7 @@ static void find_change_block_alloc(int *newval)
 		exit(-1);
 	}
 	ablock = blockstack[blockhist % BLOCK_STACK_SIZE].block;
-	if (ablock == sbd.sb_addr)
+	if (ablock == LGFS2_SB_ADDR(&sbd))
 		printf("3 (the superblock is not in the bitmap)\n");
 	else {
 		rgd = gfs2_blk2rgrpd(&sbd, ablock);
@@ -2120,7 +2121,7 @@ void gfs_log_header_in(struct gfs_log_header *head,
 {
 	struct gfs_log_header *str = lbh->iov.iov_base;
 
-	gfs2_meta_header_in(&head->lh_header, lbh);
+	gfs2_meta_header_in(&head->lh_header, lbh->b_data);
 
 	head->lh_flags = be32_to_cpu(str->lh_flags);
 	head->lh_pad = be32_to_cpu(str->lh_pad);
@@ -2158,7 +2159,7 @@ static void usage(void)
 	fprintf(stderr,"If only the device is specified, it enters into hexedit mode.\n");
 	fprintf(stderr,"identify - prints out only the block type, not the details.\n");
 	fprintf(stderr,"printsavedmeta - prints out the saved metadata blocks from a savemeta file.\n");
-	fprintf(stderr,"savemeta <file_system> <file> - save off your metadata for analysis and debugging.\n");
+	fprintf(stderr,"savemeta <file_system> <file.gz> - save off your metadata for analysis and debugging.\n");
 	fprintf(stderr,"   (The intelligent way: assume bitmap is correct).\n");
 	fprintf(stderr,"savemetaslow - save off your metadata for analysis and debugging.  The SLOW way (block by block).\n");
 	fprintf(stderr,"savergs - save off only the resource group information (rindex and rgs).\n");
@@ -2167,6 +2168,7 @@ static void usage(void)
 	fprintf(stderr,"rgflags rgnum [new flags] - print or modify flags for rg #rgnum (0 - X)\n");
 	fprintf(stderr,"rgbitmaps <rgnum> - print out the bitmaps for rgrp "
 		"rgnum.\n");
+	fprintf(stderr,"rgrepair - find and repair damaged rgrp.\n");
 	fprintf(stderr,"-V   prints version number.\n");
 	fprintf(stderr,"-c 1 selects alternate color scheme 1\n");
 	fprintf(stderr,"-d   prints details (for printing journals)\n");
@@ -2226,8 +2228,8 @@ static void usage(void)
 	fprintf(stderr,"     gfs2_edit -p quota find di /dev/x/y\n");
 	fprintf(stderr,"   To set the Resource Group flags for rg #7 to 3.\n");
 	fprintf(stderr,"     gfs2_edit rgflags 7 3 /dev/sdc2\n");
-	fprintf(stderr,"   To save off all metadata for /dev/vg/lv without compression:\n");
-	fprintf(stderr,"     gfs2_edit savemeta -z 0 /dev/vg/lv /tmp/metasave\n");
+	fprintf(stderr,"   To save off all metadata for /dev/vg/lv:\n");
+	fprintf(stderr,"     gfs2_edit savemeta /dev/vg/lv /tmp/metasave.gz\n");
 }/* usage */
 
 /**
@@ -2259,6 +2261,135 @@ static void getgziplevel(char *argv[], int *i)
 	(*i)++;
 }
 
+static int count_dinode_blks(struct rgrp_tree *rgd, int bitmap,
+			     struct gfs2_buffer_head *rbh)
+{
+	struct gfs2_buffer_head *tbh;
+	uint64_t b;
+	int dinodes = 0;
+	char *byte, cur_state, new_state;
+	int bit, off;
+
+	if (bitmap)
+		off = sizeof(struct gfs2_meta_header);
+	else
+		off = sizeof(struct gfs2_rgrp);
+
+	for (b = 0; b < rgd->bits[bitmap].bi_len << GFS2_BIT_SIZE; b++) {
+		tbh = bread(&sbd, rgd->ri.ri_data0 +
+			    rgd->bits[bitmap].bi_start + b);
+		byte = rbh->b_data + off + (b / GFS2_NBBY);
+		bit = (b % GFS2_NBBY) * GFS2_BIT_SIZE;
+		if (gfs2_check_meta(tbh, GFS2_METATYPE_DI) == 0) {
+			dinodes++;
+			new_state = GFS2_BLKST_DINODE;
+		} else {
+			new_state = GFS2_BLKST_USED;
+		}
+		cur_state = (*byte >> bit) & GFS2_BIT_MASK;
+		*byte ^= cur_state << bit;
+		*byte |= new_state << bit;
+		brelse(tbh);
+	}
+	bmodified(rbh);
+	return dinodes;
+}
+
+static int count_dinode_bits(struct gfs2_buffer_head *rbh)
+{
+	uint64_t blk;
+	struct gfs2_meta_header *mh = (struct gfs2_meta_header *)rbh->b_data;
+	char *byte;
+	int bit;
+	int dinodes = 0;
+
+	if (be32_to_cpu(mh->mh_type) == GFS2_METATYPE_RG)
+		blk = sizeof(struct gfs2_rgrp);
+	else
+		blk = sizeof(struct gfs2_meta_header);
+
+	for (; blk < sbd.bsize; blk++) {
+		byte = rbh->b_data + (blk / GFS2_NBBY);
+		bit = (blk % GFS2_NBBY) * GFS2_BIT_SIZE;
+		if (((*byte >> bit) & GFS2_BIT_MASK) == GFS2_BLKST_DINODE)
+			dinodes++;
+	}
+	return dinodes;
+}
+
+static void rg_repair(void)
+{
+	struct gfs2_buffer_head *rbh;
+	struct rgrp_tree *rgd;
+	struct osi_node *n;
+	int b;
+	int rgs_fixed = 0;
+	int dinodes_found = 0, dinodes_total = 0;
+
+	/* Walk through the resource groups saving everything within */
+	for (n = osi_first(&sbd.rgtree); n; n = osi_next(n)) {
+		rgd = (struct rgrp_tree *)n;
+		if (gfs2_rgrp_read(&sbd, rgd) == 0) { /* was read in okay */
+			gfs2_rgrp_relse(rgd);
+			continue; /* ignore it */
+		}
+		/* If we get here, it's because we have an rgrp in the rindex
+		   file that can't be read in. So attempt to repair it.
+		   If we find a damaged rgrp or bitmap, fix the metadata.
+		   Then scan all its blocks: if we find a dinode, set the
+		   repaired bitmap to GFS2_BLKST_DINODE. Set all others to
+		   GFS2_BLKST_USED so fsck can sort it out. If we set them
+		   to FREE, fsck would just nuke it all. */
+		printf("Resource group at block %llu (0x%llx) appears to be "
+		       "damaged. Attempting to fix it (in reverse order).\n",
+		       (unsigned long long)rgd->ri.ri_addr,
+		       (unsigned long long)rgd->ri.ri_addr);
+
+		for (b = rgd->ri.ri_length - 1; b >= 0; b--) {
+			int mtype = (b ? GFS2_METATYPE_RB : GFS2_METATYPE_RG);
+			struct gfs2_meta_header *mh;
+
+			printf("Bitmap #%d:", b);
+			rbh = bread(&sbd, rgd->ri.ri_addr + b);
+			if (gfs2_check_meta(rbh, mtype)) { /* wrong type */
+				printf("Damaged. Repairing...");
+				/* Fix the meta header */
+				memset(rbh->b_data, 0, sbd.bsize);
+				mh = (struct gfs2_meta_header *)rbh->b_data;
+				mh->mh_magic = cpu_to_be32(GFS2_MAGIC);
+				mh->mh_type = cpu_to_be32(mtype);
+				if (b)
+					mh->mh_format =
+						cpu_to_be32(GFS2_FORMAT_RB);
+				else
+					mh->mh_format =
+						cpu_to_be32(GFS2_FORMAT_RG);
+				bmodified(rbh);
+				/* Count the dinode blocks */
+				dinodes_found = count_dinode_blks(rgd, b, rbh);
+			} else { /* bitmap info is okay: tally it. */
+				printf("Undamaged. Analyzing...");
+				dinodes_found = count_dinode_bits(rbh);
+			}
+			printf("Dinodes found: %d\n", dinodes_found);
+			dinodes_total += dinodes_found;
+			if (b == 0) { /* rgrp itself was damaged */
+				rgd->rg.rg_dinodes = dinodes_total;
+				rgd->rg.rg_free = 0;
+			}
+			brelse(rbh);
+		}
+		rgs_fixed++;
+	}
+	if (rgs_fixed)
+		printf("%d resource groups fixed.\n"
+		       "You should run fsck.gfs2 to reconcile the bitmaps.\n",
+		       rgs_fixed);
+	else
+		printf("All resource groups are okay. No repairs needed.\n");
+	exit(0);
+}
+
 /* ------------------------------------------------------------------------ */
 /* parameterpass1 - pre-processing for command-line parameters              */
 /* ------------------------------------------------------------------------ */
@@ -2307,6 +2438,8 @@ static void parameterpass1(int argc, char *argv[], int i)
 		termlines = 0;
 	else if (!strcmp(argv[i], "rgflags"))
 		termlines = 0;
+	else if (!strcmp(argv[i], "rgrepair"))
+		termlines = 0;
 	else if (!strcmp(argv[i], "rg"))
 		termlines = 0;
 	else if (!strcasecmp(argv[i], "-x"))
@@ -2353,8 +2486,9 @@ static void process_parameters(int argc, char *argv[], int pass)
 		}
 		if (termlines || strchr(argv[i],'/')) /* if print or slash */
 			continue;
-			
-		if (!strncmp(argv[i], "journal", 7) && isdigit(argv[i][7])) {
+
+		if (!strncmp(argv[i], "journal", 7) && isdigit(argv[i][7]) &&
+		    strcmp(argv[i+1], "field")) {
 			int blk = 0;
 
 			if (i < argc - 1 && isdigit(argv[i + 1][0])) {
@@ -2483,6 +2617,8 @@ static void process_parameters(int argc, char *argv[], int pass)
 			for (bmap = 0; bmap < rgd->ri.ri_length; bmap++)
 				push_block(rgblk + bmap);
 		}
+		else if (!strcmp(argv[i], "rgrepair"))
+			rg_repair();
 		else if (!strcasecmp(argv[i], "savemeta")) {
 			getgziplevel(argv, &i);
 			savemeta(argv[i+2], 0, gziplevel);
