@@ -54,7 +54,7 @@ uint64_t find_journal_block(const char *journal, uint64_t *j_size)
 	/* read in the block */
 	jindex_bh = bread(&sbd, jindex_block);
 	/* get the dinode data from it. */
-	gfs2_dinode_in(&di, jindex_bh); /* parse disk inode to struct*/
+	gfs2_dinode_in(&di, jindex_bh->b_data);
 
 	if (!sbd.gfs1)
 		do_dinode_extended(&di, jindex_bh); /* parse dir. */
@@ -72,7 +72,7 @@ uint64_t find_journal_block(const char *journal, uint64_t *j_size)
 		if (amtread) {
 			gfs_jindex_in(&ji, jbuf);
 			jblock = ji.ji_addr;
-			*j_size = ji.ji_nsegment * 0x10;
+			*j_size = (uint64_t)ji.ji_nsegment * 0x10;
 		}
 		inode_put(&jiinode);
 	} else {
@@ -82,7 +82,7 @@ uint64_t find_journal_block(const char *journal, uint64_t *j_size)
 			return 0;
 		jblock = indirect->ii[0].dirent[journal_num + 2].block;
 		j_bh = bread(&sbd, jblock);
-		gfs2_dinode_in(&jdi, j_bh);/* parse dinode to struct */
+		gfs2_dinode_in(&jdi, j_bh->b_data);
 		*j_size = jdi.di_size;
 		brelse(j_bh);
 	}
@@ -317,7 +317,7 @@ static int is_wrap_pt(char *buf, uint64_t *highest_seq)
 			seq = lh.lh_sequence;
 		} else {
 			struct gfs2_log_header lh;
-			gfs2_log_header_in(&lh, &tbh);
+			gfs2_log_header_in(&lh, buf);
 			seq = lh.lh_sequence;
 		}
 		if (seq < *highest_seq)
@@ -364,7 +364,7 @@ static uint64_t find_wrap_pt(struct gfs2_inode *ji, char *jbuf, uint64_t jblock,
  * process_ld - process a log descriptor
  */
 static int process_ld(uint64_t abs_block, uint64_t wrappt, uint64_t j_size,
-		      uint64_t jb, struct gfs2_buffer_head *dummy_bh, int tblk,
+		      uint64_t jb, char *buf, int tblk,
 		      uint64_t *tblk_off, uint64_t bitblk,
 		      struct rgrp_tree *rgd, int *prnt, uint64_t *bblk_off)
 {
@@ -383,15 +383,12 @@ static int process_ld(uint64_t abs_block, uint64_t wrappt, uint64_t j_size,
 		{"Metadata", "Unlinked inode", "Dealloc inode",
 		 "Quota", "Final Entry", "Unknown"}};
 
-	gfs2_log_descriptor_in(&ld, dummy_bh);
+	gfs2_log_descriptor_in(&ld, buf);
 	if (sbd.gfs1)
-		b = (uint64_t *)(dummy_bh->b_data +
-				 sizeof(struct gfs_log_descriptor));
+		b = (uint64_t *)(buf + sizeof(struct gfs_log_descriptor));
 	else
-		b = (uint64_t *)(dummy_bh->b_data +
-				 sizeof(struct gfs2_log_descriptor));
-	*prnt = ld_is_pertinent(b, (dummy_bh->b_data + sbd.bsize), tblk, rgd,
-			       bitblk);
+		b = (uint64_t *)(buf + sizeof(struct gfs2_log_descriptor));
+	*prnt = ld_is_pertinent(b, (buf + sbd.bsize), tblk, rgd, bitblk);
 
 	if (*prnt) {
 		print_gfs2("0x%"PRIx64" (j+%4"PRIx64"): Log descriptor, ",
@@ -412,9 +409,9 @@ static int process_ld(uint64_t abs_block, uint64_t wrappt, uint64_t j_size,
 	if (ld.ld_type == GFS2_LOG_DESC_METADATA ||
 	    ld.ld_type == GFS_LOG_DESC_METADATA)
 		is_meta_ld = 1;
-	ld_blocks -= print_ld_blks(b, (dummy_bh->b_data + sbd.bsize), line,
-				   tblk, tblk_off, bitblk, rgd, abs_block,
-				   *prnt, bblk_off, is_meta_ld);
+	ld_blocks -= print_ld_blks(b, (buf + sbd.bsize), line, tblk, tblk_off,
+	                           bitblk, rgd, abs_block, *prnt, bblk_off,
+	                           is_meta_ld);
 
 	return ld_blocks;
 }
@@ -533,7 +530,7 @@ void dump_journal(const char *journal, int tblk)
 				   rgd->ri.ri_length);
 			o = tblk - rgd->ri.ri_data0;
 			if (o >= (rgd->bits->bi_start +
-				  rgd->bits->bi_len) * GFS2_NBBY)
+				  rgd->bits->bi_len) * (uint64_t)GFS2_NBBY)
 				o += (sizeof(struct gfs2_rgrp) -
 				      sizeof(struct gfs2_meta_header))
 					* GFS2_NBBY;
@@ -575,7 +572,7 @@ void dump_journal(const char *journal, int tblk)
 		block_type = get_block_type(&dummy_bh, NULL);
 		if (block_type == GFS2_METATYPE_LD) {
 			ld_blocks = process_ld(abs_block, wrappt, j_size, jb,
-					       &dummy_bh, tblk, &tblk_off,
+					       dummy_bh.b_data, tblk, &tblk_off,
 					       bitblk, rgd, &is_pertinent,
 					       &bblk_off);
 			offset_from_ld = 0;
@@ -596,7 +593,7 @@ void dump_journal(const char *journal, int tblk)
 					   lh1.lh_first, lh1.lh_tail,
 					   lh1.lh_last_dump);
 			} else {
-				gfs2_log_header_in(&lh, &dummy_bh);
+				gfs2_log_header_in(&lh, dummy_bh.b_data);
 				check_journal_wrap(lh.lh_sequence,
 						   &highest_seq);
 				print_gfs2("0x%"PRIx64" (j+%4"PRIx64"): Log header: Seq"
@@ -645,7 +642,8 @@ void dump_journal(const char *journal, int tblk)
 			block = saveblk;
 		}
 	}
-	inode_put(&j_inode);
+	if (j_inode != NULL)
+		inode_put(&j_inode);
 	brelse(j_bh);
 	blockhist = -1; /* So we don't print anything else */
 	free(jbuf);
