@@ -16,7 +16,7 @@
 #include "metawalk.h"
 #include "util.h"
 
-#define JOURNAL_NAME_SIZE 16
+#define JOURNAL_NAME_SIZE 18
 #define JOURNAL_SEQ_TOLERANCE 10
 
 unsigned int sd_found_jblocks = 0, sd_replayed_jblocks = 0;
@@ -114,7 +114,7 @@ static void refresh_rgrp(struct gfs2_sbd *sdp, struct rgrp_tree *rgd,
 				gfs_rgrp_in((struct gfs_rgrp *)&rgd->rg,
 					    rgd->bits[0].bi_bh);
 			else
-				gfs2_rgrp_in(&rgd->rg, rgd->bits[0].bi_bh);
+				gfs2_rgrp_in(&rgd->rg, rgd->bits[0].bi_bh->b_data);
 		}
 		break;
 	}
@@ -406,7 +406,6 @@ static int check_journal_seq_no(struct gfs2_inode *ip, int fix)
 	uint64_t highest_seq = 0, lowest_seq = 0, prev_seq = 0;
 	int new = 0;
 	uint64_t dblock;
-	uint32_t extlen;
 	struct gfs2_buffer_head *bh;
 	int seq_errors = 0;
 
@@ -443,9 +442,10 @@ static int check_journal_seq_no(struct gfs2_inode *ip, int fix)
 		lh.lh_sequence = highest_seq;
 		prev_seq = lh.lh_sequence;
 		log_warn( _("Renumbering it as 0x%llx\n"), lh.lh_sequence);
-		block_map(ip, blk, &new, &dblock, &extlen, FALSE);
+		block_map(ip, blk, &new, &dblock, NULL, FALSE);
 		bh = bread(ip->i_sbd, dblock);
-		gfs2_log_header_out(&lh, bh);
+		gfs2_log_header_out(&lh, bh->b_data);
+		bmodified(bh);
 		brelse(bh);
 	}
 	if (seq_errors && fix) {
@@ -626,7 +626,7 @@ reinit:
  * through initialization properly yet. */
 static int rangecheck_jblock(struct gfs2_inode *ip, uint64_t block)
 {
-	if((block > ip->i_sbd->fssize) || (block <= ip->i_sbd->sb_addr)) {
+	if((block > ip->i_sbd->fssize) || (block <= LGFS2_SB_ADDR(ip->i_sbd))) {
 		log_info( _("Bad block pointer (out of range) found in "
 			    "journal inode %lld (0x%llx).\n"),
 			  (unsigned long long)ip->i_di.di_num.no_addr,
