@@ -602,7 +602,7 @@ static int check_data(struct gfs2_inode *ip, uint64_t metablock,
 			return 1;
 		case GFS2_BLKST_USED: /* tough decision: May be data or meta */
 			bh = bread(ip->i_sbd, block);
-			gfs2_meta_header_in(&mh, bh);
+			gfs2_meta_header_in(&mh, bh->b_data);
 			brelse(bh);
 			if (mh.mh_magic == GFS2_MAGIC &&
 			    mh.mh_type >= GFS2_METATYPE_RG &&
@@ -1569,7 +1569,7 @@ static void check_i_goal(struct gfs2_sbd *sdp, struct gfs2_inode *ip)
 	if (sdp->gfs1 || ip->i_di.di_flags & GFS2_DIF_SYSTEM)
 		return;
 
-	if (ip->i_di.di_goal_meta <= sdp->sb_addr ||
+	if (ip->i_di.di_goal_meta <= LGFS2_SB_ADDR(sdp) ||
 	    ip->i_di.di_goal_meta > sdp->fssize) {
 		log_err(_("Inode #%llu (0x%llx): Bad allocation goal block "
 			  "found: %llu (0x%llx)\n"),
@@ -2123,7 +2123,7 @@ static void enomem(uint64_t addl_mem_needed)
 	log_crit( _("This system doesn't have enough memory and swap space to fsck this file system.\n"));
 	log_crit( _("Additional memory needed is approximately: %lluMB\n"),
 		  (unsigned long long)(addl_mem_needed / 1048576ULL));
-	log_crit( _("Please increase your swap space by that amount and run gfs2_fsck again.\n"));
+	log_crit( _("Please increase your swap space by that amount and run fsck.gfs2 again.\n"));
 }
 
 /**
