@@ -39,7 +39,7 @@ struct gfs2_inode *lgfs2_inode_get(struct gfs2_sbd *sdp, struct gfs2_buffer_head
 	if (ip == NULL) {
 		return NULL;
 	}
-	gfs2_dinode_in(&ip->i_di, bh);
+	gfs2_dinode_in(&ip->i_di, bh->b_data);
 	ip->i_bh = bh;
 	ip->i_sbd = sdp;
 	return ip;
@@ -95,7 +95,7 @@ void inode_put(struct gfs2_inode **ip_in)
 	struct gfs2_sbd *sdp = ip->i_sbd;
 
 	if (ip->i_bh->b_modified) {
-		gfs2_dinode_out(&ip->i_di, ip->i_bh);
+		gfs2_dinode_out(&ip->i_di, ip->i_bh->b_data);
 		if (!ip->bh_owned && is_system_inode(sdp, block))
 			fprintf(stderr, "Warning: Change made to inode "
 				"were discarded.\n");
@@ -158,8 +158,8 @@ static int blk_alloc_in_rg(struct gfs2_sbd *sdp, unsigned state, struct rgrp_tre
 	if (sdp->gfs1)
 		gfs_rgrp_out((struct gfs_rgrp *)&rgd->rg, rgd->bits[0].bi_bh);
 	else
-		gfs2_rgrp_out_bh(&rgd->rg, rgd->bits[0].bi_bh);
-
+		gfs2_rgrp_out(&rgd->rg, rgd->bits[0].bi_bh->b_data);
+	bmodified(rgd->bits[0].bi_bh);
 	sdp->blks_alloced++;
 	return 0;
 }
@@ -255,12 +255,13 @@ void unstuff_dinode(struct gfs2_inode *ip)
 			mh.mh_magic = GFS2_MAGIC;
 			mh.mh_type = GFS2_METATYPE_JD;
 			mh.mh_format = GFS2_FORMAT_JD;
-			gfs2_meta_header_out_bh(&mh, bh);
+			gfs2_meta_header_out(&mh, bh->b_data);
 
 			buffer_copy_tail(sdp, bh,
 					 sizeof(struct gfs2_meta_header),
 					 ip->i_bh, sizeof(struct gfs2_dinode));
 
+			bmodified(bh);
 			brelse(bh);
 		} else {
 			bh = bget(sdp, block);
@@ -410,10 +411,11 @@ void build_height(struct gfs2_inode *ip, int height)
 			mh.mh_magic = GFS2_MAGIC;
 			mh.mh_type = GFS2_METATYPE_IN;
 			mh.mh_format = GFS2_FORMAT_IN;
-			gfs2_meta_header_out_bh(&mh, bh);
+			gfs2_meta_header_out(&mh, bh->b_data);
 			buffer_copy_tail(sdp, bh,
 					 sizeof(struct gfs2_meta_header),
 					 ip->i_bh, sizeof(struct gfs2_dinode));
+			bmodified(bh);
 			brelse(bh);
 		}
 
@@ -521,7 +523,8 @@ void block_map(struct gfs2_inode *ip, uint64_t lblock, int *new,
 			mh.mh_magic = GFS2_MAGIC;
 			mh.mh_type = GFS2_METATYPE_IN;
 			mh.mh_format = GFS2_FORMAT_IN;
-			gfs2_meta_header_out_bh(&mh, bh);
+			gfs2_meta_header_out(&mh, bh->b_data);
+			bmodified(bh);
 		} else {
 			if (*dblock == ip->i_di.di_num.no_addr)
 				bh = ip->i_bh;
@@ -712,7 +715,8 @@ int __gfs2_writei(struct gfs2_inode *ip, void *buf,
 				mh.mh_magic = GFS2_MAGIC;
 				mh.mh_type = GFS2_METATYPE_JD;
 				mh.mh_format = GFS2_FORMAT_JD;
-				gfs2_meta_header_out_bh(&mh, bh);
+				gfs2_meta_header_out(&mh, bh->b_data);
+				bmodified(bh);
 			}
 		} else {
 			if (dblock == ip->i_di.di_num.no_addr)
@@ -938,7 +942,8 @@ void dir_split_leaf(struct gfs2_inode *dip, uint32_t start, uint64_t leaf_no,
 		mh.mh_magic = GFS2_MAGIC;
 		mh.mh_type = GFS2_METATYPE_LF;
 		mh.mh_format = GFS2_FORMAT_LF;
-		gfs2_meta_header_out_bh(&mh, nbh);
+		gfs2_meta_header_out(&mh, nbh->b_data);
+		bmodified(nbh);
 		buffer_clear_tail(dip->i_sbd, nbh,
 				  sizeof(struct gfs2_meta_header));
 	}
@@ -1218,7 +1223,8 @@ restart:
 				mh.mh_magic = GFS2_MAGIC;
 				mh.mh_type = GFS2_METATYPE_LF;
 				mh.mh_format = GFS2_FORMAT_LF;
-				gfs2_meta_header_out_bh(&mh, nbh);
+				gfs2_meta_header_out(&mh, nbh->b_data);
+				bmodified(nbh);
 
 				leaf->lf_next = cpu_to_be64(bn);
 
@@ -1272,7 +1278,8 @@ static void dir_make_exhash(struct gfs2_inode *dip)
 		mh.mh_magic = GFS2_MAGIC;
 		mh.mh_type = GFS2_METATYPE_LF;
 		mh.mh_format = GFS2_FORMAT_LF;
-		gfs2_meta_header_out_bh(&mh, bh);
+		gfs2_meta_header_out(&mh, bh->b_data);
+		bmodified(bh);
 	}
 
 	leaf = (struct gfs2_leaf *)bh->b_data;
@@ -1317,7 +1324,7 @@ static void dir_make_exhash(struct gfs2_inode *dip)
 	for (x = sdp->sd_hash_ptrs, y = -1; x; x >>= 1, y++) ;
 	dip->i_di.di_depth = y;
 
-	gfs2_dinode_out(&dip->i_di, dip->i_bh);
+	gfs2_dinode_out(&dip->i_di, dip->i_bh->b_data);
 	bwrite(dip->i_bh);
 }
 
@@ -1423,9 +1430,8 @@ static int __init_dinode(struct gfs2_sbd *sdp, struct gfs2_buffer_head **bhp, st
 		di.di_payload_format = GFS2_FORMAT_DE;
 		di.di_entries = 2;
 	}
-
-	gfs2_dinode_out(&di, bh);
-
+	gfs2_dinode_out(&di, bh->b_data);
+	bmodified(bh);
 	return 0;
 }
 
@@ -1487,13 +1493,13 @@ int lgfs2_write_filemeta(struct gfs2_inode *ip)
 			char *start = bh->b_data;
 			if (height == 0) {
 				start += sizeof(struct gfs2_dinode);
-				gfs2_dinode_out(&ip->i_di, bh);
+				gfs2_dinode_out(&ip->i_di, bh->b_data);
 			} else {
 				start += sizeof(struct gfs2_meta_header);
 				gfs2_meta_header_out(&mh, bh->b_data);
 			}
 			lgfs2_fill_indir(start, bh->b_data + sdp->bsize, ptr0, ptrs, &p);
-			if(bwrite(bh)) {
+			if (bwrite(bh)) {
 				free(bh);
 				return 1;
 			}
@@ -1924,7 +1930,8 @@ void gfs2_free_block(struct gfs2_sbd *sdp, uint64_t block)
 		if (sdp->gfs1)
 			gfs_rgrp_out((struct gfs_rgrp *)&rgd->rg, rgd->bits[0].bi_bh);
 		else
-			gfs2_rgrp_out_bh(&rgd->rg, rgd->bits[0].bi_bh);
+			gfs2_rgrp_out(&rgd->rg, rgd->bits[0].bi_bh->b_data);
+		bmodified(rgd->bits[0].bi_bh);
 		sdp->blks_alloced--;
 	}
 }
@@ -1993,7 +2000,8 @@ int gfs2_freedi(struct gfs2_sbd *sdp, uint64_t diblock)
 	if (sdp->gfs1)
 		gfs_rgrp_out((struct gfs_rgrp *)&rgd->rg, rgd->bits[0].bi_bh);
 	else
-		gfs2_rgrp_out_bh(&rgd->rg, rgd->bits[0].bi_bh);
+		gfs2_rgrp_out(&rgd->rg, rgd->bits[0].bi_bh->b_data);
+	bmodified(rgd->bits[0].bi_bh);
 	sdp->dinodes_alloced--;
 	return 0;
 }
