@@ -6,9 +6,11 @@
 #include <stdint.h>
 #include <inttypes.h>
 #include <ctype.h>
-
 #include <linux/types.h>
 #include "libgfs2.h"
+#ifdef GFS2_HAS_UUID
+#include <uuid.h>
+#endif
 
 #define pv(struct, member, fmt, fmt2) do {				\
 		print_it("  "#member, fmt, fmt2, struct->member);	\
@@ -30,11 +32,11 @@
 /*
  * gfs2_xxx_in - read in an xxx struct
  * first arg: the cpu-order structure
- * bh: the disk-order buffer_head
+ * buf: the disk-order block data
  *
  * gfs2_xxx_out - write out an xxx struct
  * first arg: the cpu-order structure
- * bh: the disk-order buffer_head
+ * buf: the disk-order block data
  *
  * gfs2_xxx_print - print out an xxx struct
  * first arg: the cpu-order structure
@@ -62,10 +64,9 @@ void gfs2_inum_print(const struct gfs2_inum *no)
 	pv(no, no_addr, "%llu", "0x%llx");
 }
 
-void gfs2_meta_header_in(struct gfs2_meta_header *mh,
-			 struct gfs2_buffer_head *bh)
+void gfs2_meta_header_in(struct gfs2_meta_header *mh, char *buf)
 {
-	struct gfs2_meta_header *str = (struct gfs2_meta_header *)bh->b_data;
+	struct gfs2_meta_header *str = (struct gfs2_meta_header *)buf;
 
 	CPIN_32(mh, str, mh_magic);
 	CPIN_32(mh, str, mh_type);
@@ -83,13 +84,6 @@ void gfs2_meta_header_out(const struct gfs2_meta_header *mh, char *buf)
 	str->__pad1 = 0;
 }
 
-void gfs2_meta_header_out_bh(const struct gfs2_meta_header *mh,
-                             struct gfs2_buffer_head *bh)
-{
-	gfs2_meta_header_out(mh, bh->iov.iov_base);
-	bmodified(bh);
-}
-
 void gfs2_meta_header_print(const struct gfs2_meta_header *mh)
 {
 	pv(mh, mh_magic, "0x%08X", NULL);
@@ -97,11 +91,11 @@ void gfs2_meta_header_print(const struct gfs2_meta_header *mh)
 	pv(mh, mh_format, "%u", "0x%x");
 }
 
-void gfs2_sb_in(struct gfs2_sb *sb, struct gfs2_buffer_head *bh)
+void gfs2_sb_in(struct gfs2_sb *sb, char *buf)
 {
-	struct gfs2_sb *str = (struct gfs2_sb *)bh->b_data;
+	struct gfs2_sb *str = (struct gfs2_sb *)buf;
 
-	gfs2_meta_header_in(&sb->sb_header, bh);
+	gfs2_meta_header_in(&sb->sb_header, buf);
 
 	CPIN_32(sb, str, sb_fs_format);
 	CPIN_32(sb, str, sb_multihost_format);
@@ -151,32 +145,6 @@ void gfs2_sb_out(const struct gfs2_sb *sb, char *buf)
 #endif
 }
 
-const char *str_uuid(const unsigned char *uuid)
-{
-	static char str[64];
-	char *ch;
-	int i;
-
-	memset(str, 0, sizeof(str));
-	ch = str;
-	for (i = 0; i < 16; i++) {
-		sprintf(ch, "%02x", uuid[i]);
-		ch += 2;
-		if ((i == 3) || (i == 5) || (i == 7) || (i == 9)) {
-			*ch = '-';
-			ch++;
-		}
-	}
-	return str;
-}
-
-#ifdef GFS2_HAS_UUID
-void gfs2_print_uuid(const unsigned char *uuid)
-{
-	print_it("  uuid", "%s", NULL, str_uuid(uuid));
-}
-#endif
-
 void gfs2_sb_print(const struct gfs2_sb *sb)
 {
 	gfs2_meta_header_print(&sb->sb_header);
@@ -194,7 +162,12 @@ void gfs2_sb_print(const struct gfs2_sb *sb)
 	pv(sb, sb_locktable, "%s", NULL);
 
 #ifdef GFS2_HAS_UUID
-	gfs2_print_uuid(sb->sb_uuid);
+	{
+	char readable_uuid[36+1];
+
+	uuid_unparse(sb->sb_uuid, readable_uuid);
+	print_it("  uuid", "%36s", NULL, readable_uuid);
+	}
 #endif
 }
 
@@ -204,13 +177,11 @@ void gfs2_rindex_in(struct gfs2_rindex *ri, char *buf)
 
 	CPIN_64(ri, str, ri_addr);
 	CPIN_32(ri, str, ri_length);
-
+	CPIN_32(ri, str, __pad);
 	CPIN_64(ri, str, ri_data0);
 	CPIN_32(ri, str, ri_data);
-
 	CPIN_32(ri, str, ri_bitbytes);
-
-	CPIN_08(ri, str, ri_reserved, 64);
+	CPIN_08(ri, str, ri_reserved, sizeof(ri->ri_reserved));
 }
 
 void gfs2_rindex_out(const struct gfs2_rindex *ri, char *buf)
@@ -226,7 +197,7 @@ void gfs2_rindex_out(const struct gfs2_rindex *ri, char *buf)
 
 	CPOUT_32(ri, str, ri_bitbytes);
 
-	CPOUT_08(ri, str, ri_reserved, 64);
+	CPOUT_08(ri, str, ri_reserved, sizeof(ri->ri_reserved));
 }
 
 void gfs2_rindex_print(const struct gfs2_rindex *ri)
@@ -240,16 +211,27 @@ void gfs2_rindex_print(const struct gfs2_rindex *ri)
 	pv(ri, ri_bitbytes, "%u", "0x%x");
 }
 
-void gfs2_rgrp_in(struct gfs2_rgrp *rg, struct gfs2_buffer_head *bh)
+void gfs2_rgrp_in(struct gfs2_rgrp *rg, char *buf)
 {
-	struct gfs2_rgrp *str = (struct gfs2_rgrp *)bh->b_data;
+	struct gfs2_rgrp *str = (struct gfs2_rgrp *)buf;
 
-	gfs2_meta_header_in(&rg->rg_header, bh);
+	gfs2_meta_header_in(&rg->rg_header, buf);
 	CPIN_32(rg, str, rg_flags);
 	CPIN_32(rg, str, rg_free);
 	CPIN_32(rg, str, rg_dinodes);
-
-	CPIN_08(rg, str, rg_reserved, 80);
+#ifdef GFS2_HAS_RG_SKIP
+	CPIN_32(rg, str, rg_skip);
+#else
+	CPIN_32(rg, str, __pad);
+#endif
+	CPIN_64(rg, str, rg_igeneration);
+#ifdef GFS2_HAS_RG_RI_FIELDS
+	CPIN_64(rg, str, rg_data0);
+	CPIN_32(rg, str, rg_data);
+	CPIN_32(rg, str, rg_bitbytes);
+	CPIN_32(rg, str, rg_crc);
+#endif
+	CPIN_08(rg, str, rg_reserved, sizeof(rg->rg_reserved));
 }
 
 void gfs2_rgrp_out(const struct gfs2_rgrp *rg, char *buf)
@@ -260,14 +242,21 @@ void gfs2_rgrp_out(const struct gfs2_rgrp *rg, char *buf)
 	CPOUT_32(rg, str, rg_flags);
 	CPOUT_32(rg, str, rg_free);
 	CPOUT_32(rg, str, rg_dinodes);
-
-	CPOUT_08(rg, str, rg_reserved, 80);
-}
-
-void gfs2_rgrp_out_bh(const struct gfs2_rgrp *rg, struct gfs2_buffer_head *bh)
-{
-	gfs2_rgrp_out(rg, bh->iov.iov_base);
-	bmodified(bh);
+#ifdef GFS2_HAS_RG_SKIP
+	CPOUT_32(rg, str, rg_skip);
+#else
+	CPOUT_32(rg, str, __pad);
+#endif
+	CPOUT_64(rg, str, rg_igeneration);
+#ifdef GFS2_HAS_RG_RI_FIELDS
+	CPOUT_64(rg, str, rg_data0);
+	CPOUT_32(rg, str, rg_data);
+	CPOUT_32(rg, str, rg_bitbytes);
+	CPOUT_08(rg, str, rg_reserved, sizeof(rg->rg_reserved));
+	lgfs2_rgrp_crc_set(buf);
+#else
+	CPOUT_08(rg, str, rg_reserved, sizeof(rg->rg_reserved));
+#endif
 }
 
 void gfs2_rgrp_print(const struct gfs2_rgrp *rg)
@@ -276,6 +265,18 @@ void gfs2_rgrp_print(const struct gfs2_rgrp *rg)
 	pv(rg, rg_flags, "%u", "0x%x");
 	pv(rg, rg_free, "%u", "0x%x");
 	pv(rg, rg_dinodes, "%u", "0x%x");
+#ifdef GFS2_HAS_RG_SKIP
+	pv(rg, rg_skip, "%u", "0x%x");
+#else
+	pv(rg, __pad, "%u", "0x%x");
+#endif
+	pv(rg, rg_igeneration, "%llu", "0x%llx");
+#ifdef GFS2_HAS_RG_RI_FIELDS
+	pv(rg, rg_data0, "%llu", "0x%llx");
+	pv(rg, rg_data, "%u", "0x%x");
+	pv(rg, rg_bitbytes, "%u", "0x%x");
+	pv(rg, rg_crc, "%u", "0x%x");
+#endif
 }
 
 void gfs2_quota_in(struct gfs2_quota *qu, char *buf)
@@ -305,11 +306,11 @@ void gfs2_quota_print(const struct gfs2_quota *qu)
 	pv(qu, qu_value, "%lld", "0x%llx");
 }
 
-void gfs2_dinode_in(struct gfs2_dinode *di, struct gfs2_buffer_head *bh)
+void gfs2_dinode_in(struct gfs2_dinode *di, char *buf)
 {
-	struct gfs2_dinode *str = (struct gfs2_dinode *)bh->b_data;
+	struct gfs2_dinode *str = (struct gfs2_dinode *)buf;
 
-	gfs2_meta_header_in(&di->di_header, bh);
+	gfs2_meta_header_in(&di->di_header, buf);
 	gfs2_inum_in(&di->di_num, (char *)&str->di_num);
 
 	CPIN_32(di, str, di_mode);
@@ -340,11 +341,11 @@ void gfs2_dinode_in(struct gfs2_dinode *di, struct gfs2_buffer_head *bh)
 	CPIN_08(di, str, di_reserved, 32);
 }
 
-void gfs2_dinode_out(struct gfs2_dinode *di, struct gfs2_buffer_head *bh)
+void gfs2_dinode_out(struct gfs2_dinode *di, char *buf)
 {
-	struct gfs2_dinode *str = (struct gfs2_dinode *)bh->b_data;
+	struct gfs2_dinode *str = (struct gfs2_dinode *)buf;
 
-	gfs2_meta_header_out_bh(&di->di_header, bh);
+	gfs2_meta_header_out(&di->di_header, buf);
 	gfs2_inum_out(&di->di_num, (char *)&str->di_num);
 
 	CPOUT_32(di, str, di_mode);
@@ -373,7 +374,6 @@ void gfs2_dinode_out(struct gfs2_dinode *di, struct gfs2_buffer_head *bh)
 	CPOUT_64(di, str, di_eattr);
 
 	CPOUT_08(di, str, di_reserved, 32);
-	bmodified(bh);
 }
 
 void gfs2_dinode_print(const struct gfs2_dinode *di)
@@ -450,11 +450,11 @@ void gfs2_dirent_out(struct gfs2_dirent *de, char *buf)
 #endif /* GFS2_HAS_DE_RAHEAD */
 }
 
-void gfs2_leaf_in(struct gfs2_leaf *lf, struct gfs2_buffer_head *bh)
+void gfs2_leaf_in(struct gfs2_leaf *lf, char *buf)
 {
-	struct gfs2_leaf *str = (struct gfs2_leaf *)bh->b_data;
+	struct gfs2_leaf *str = (struct gfs2_leaf *)buf;
 
-	gfs2_meta_header_in(&lf->lf_header, bh);
+	gfs2_meta_header_in(&lf->lf_header, buf);
 	CPIN_16(lf, str, lf_depth);
 	CPIN_16(lf, str, lf_entries);
 	CPIN_32(lf, str, lf_dirent_format);
@@ -470,11 +470,11 @@ void gfs2_leaf_in(struct gfs2_leaf *lf, struct gfs2_buffer_head *bh)
 #endif
 }
 
-void gfs2_leaf_out(struct gfs2_leaf *lf, struct gfs2_buffer_head *bh)
+void gfs2_leaf_out(struct gfs2_leaf *lf, char *buf)
 {
-	struct gfs2_leaf *str = (struct gfs2_leaf *)bh->b_data;
+	struct gfs2_leaf *str = (struct gfs2_leaf *)buf;
 
-	gfs2_meta_header_out_bh(&lf->lf_header, bh);
+	gfs2_meta_header_out(&lf->lf_header, buf);
 	CPOUT_16(lf, str, lf_depth);
 	CPOUT_16(lf, str, lf_entries);
 	CPOUT_32(lf, str, lf_dirent_format);
@@ -488,7 +488,6 @@ void gfs2_leaf_out(struct gfs2_leaf *lf, struct gfs2_buffer_head *bh)
 #else
 	CPOUT_08(lf, str, lf_reserved, 64);
 #endif
-	bmodified(bh);
 }
 
 void gfs2_leaf_print(const struct gfs2_leaf *lf)
@@ -534,12 +533,11 @@ void gfs2_ea_header_print(const struct gfs2_ea_header *ea, char *name)
 	print_it("  name", "%s", NULL, buf);
 }
 
-void gfs2_log_header_in(struct gfs2_log_header *lh,
-			struct gfs2_buffer_head *bh)
+void gfs2_log_header_v1_in(struct gfs2_log_header *lh, char *buf)
 {
-	struct gfs2_log_header *str = (struct gfs2_log_header *)bh->b_data;
+	struct gfs2_log_header *str = (struct gfs2_log_header *)buf;
 
-	gfs2_meta_header_in(&lh->lh_header, bh);
+	gfs2_meta_header_in(&lh->lh_header, buf);
 	CPIN_64(lh, str, lh_sequence);
 	CPIN_32(lh, str, lh_flags);
 	CPIN_32(lh, str, lh_tail);
@@ -547,21 +545,61 @@ void gfs2_log_header_in(struct gfs2_log_header *lh,
 	CPIN_32(lh, str, lh_hash);
 }
 
-void gfs2_log_header_out(struct gfs2_log_header *lh,
-			 struct gfs2_buffer_head *bh)
+void gfs2_log_header_in(struct gfs2_log_header *lh, char *buf)
 {
-	struct gfs2_log_header *str = (struct gfs2_log_header *)bh->b_data;
+	gfs2_log_header_v1_in(lh, buf);
+#ifdef GFS2_HAS_LH_V2
+	{
+	struct gfs2_log_header *str = (struct gfs2_log_header *)buf;
 
-	gfs2_meta_header_out_bh(&lh->lh_header, bh);
+	CPIN_32(lh, str, lh_crc);
+	CPIN_32(lh, str, lh_nsec);
+	CPIN_64(lh, str, lh_sec);
+	CPIN_64(lh, str, lh_addr);
+	CPIN_64(lh, str, lh_jinode);
+	CPIN_64(lh, str, lh_statfs_addr);
+	CPIN_64(lh, str, lh_quota_addr);
+	CPIN_64(lh, str, lh_local_total);
+	CPIN_64(lh, str, lh_local_free);
+	CPIN_64(lh, str, lh_local_dinodes);
+	}
+#endif
+}
+
+void gfs2_log_header_v1_out(struct gfs2_log_header *lh, char *buf)
+{
+	struct gfs2_log_header *str = (struct gfs2_log_header *)buf;
+
+	gfs2_meta_header_out(&lh->lh_header, buf);
 	CPOUT_64(lh, str, lh_sequence);
 	CPOUT_32(lh, str, lh_flags);
 	CPOUT_32(lh, str, lh_tail);
 	CPOUT_32(lh, str, lh_blkno);
 	CPOUT_32(lh, str, lh_hash);
-	bmodified(bh);
 }
 
-void gfs2_log_header_print(const struct gfs2_log_header *lh)
+void gfs2_log_header_out(struct gfs2_log_header *lh, char *buf)
+{
+	gfs2_log_header_v1_out(lh, buf);
+#ifdef GFS2_HAS_LH_V2
+	{
+	struct gfs2_log_header *str = (struct gfs2_log_header *)buf;
+
+	CPOUT_32(lh, str, lh_crc);
+	CPOUT_32(lh, str, lh_nsec);
+	CPOUT_64(lh, str, lh_sec);
+	CPOUT_64(lh, str, lh_addr);
+	CPOUT_64(lh, str, lh_jinode);
+	CPOUT_64(lh, str, lh_statfs_addr);
+	CPOUT_64(lh, str, lh_quota_addr);
+	CPOUT_64(lh, str, lh_local_total);
+	CPOUT_64(lh, str, lh_local_free);
+	CPOUT_64(lh, str, lh_local_dinodes);
+	}
+#endif
+}
+
+void gfs2_log_header_v1_print(const struct gfs2_log_header *lh)
 {
 	gfs2_meta_header_print(&lh->lh_header);
 	pv(lh, lh_sequence, "%llu", "0x%llx");
@@ -571,12 +609,28 @@ void gfs2_log_header_print(const struct gfs2_log_header *lh)
 	pv(lh, lh_hash, "0x%.8X", NULL);
 }
 
-void gfs2_log_descriptor_in(struct gfs2_log_descriptor *ld,
-			    struct gfs2_buffer_head *bh)
+void gfs2_log_header_print(const struct gfs2_log_header *lh)
 {
-	struct gfs2_log_descriptor *str = (struct gfs2_log_descriptor *)bh->b_data;
+	gfs2_log_header_v1_print(lh);
+#ifdef GFS2_HAS_LH_V2
+	pv(lh, lh_crc, "0x%.8X", NULL);
+	pv(lh, lh_nsec, "%u", "0x%x");
+	pv(lh, lh_sec, "%llu", "0x%llx");
+	pv(lh, lh_addr, "%llu", "0x%llx");
+	pv(lh, lh_jinode, "%llu", "0x%llx");
+	pv(lh, lh_statfs_addr, "%llu", "0x%llx");
+	pv(lh, lh_quota_addr, "%llu", "0x%llx");
+	pv(lh, lh_local_total, "%lld", "0x%llx");
+	pv(lh, lh_local_free, "%lld", "0x%llx");
+	pv(lh, lh_local_dinodes, "%lld", "0x%llx");
+#endif
+}
 
-	gfs2_meta_header_in(&ld->ld_header, bh);
+void gfs2_log_descriptor_in(struct gfs2_log_descriptor *ld, char *buf)
+{
+	struct gfs2_log_descriptor *str = (struct gfs2_log_descriptor *)buf;
+
+	gfs2_meta_header_in(&ld->ld_header, buf);
 	CPIN_32(ld, str, ld_type);
 	CPIN_32(ld, str, ld_length);
 	CPIN_32(ld, str, ld_data1);
@@ -585,19 +639,17 @@ void gfs2_log_descriptor_in(struct gfs2_log_descriptor *ld,
 	CPIN_08(ld, str, ld_reserved, 32);
 }
 
-void gfs2_log_descriptor_out(struct gfs2_log_descriptor *ld,
-			     struct gfs2_buffer_head *bh)
+void gfs2_log_descriptor_out(struct gfs2_log_descriptor *ld, char *buf)
 {
-	struct gfs2_log_descriptor *str = (struct gfs2_log_descriptor *)bh->b_data;
+	struct gfs2_log_descriptor *str = (struct gfs2_log_descriptor *)buf;
 
-	gfs2_meta_header_out_bh(&ld->ld_header, bh);
+	gfs2_meta_header_out(&ld->ld_header, buf);
 	CPOUT_32(ld, str, ld_type);
 	CPOUT_32(ld, str, ld_length);
 	CPOUT_32(ld, str, ld_data1);
 	CPOUT_32(ld, str, ld_data2);
 
 	CPOUT_08(ld, str, ld_reserved, 32);
-	bmodified(bh);
 }
 
 void gfs2_log_descriptor_print(const struct gfs2_log_descriptor *ld)
@@ -634,27 +686,24 @@ void gfs2_statfs_change_print(const struct gfs2_statfs_change *sc)
 	pv(sc, sc_dinodes, "%lld", "0x%llx");
 }
 
-void gfs2_quota_change_in(struct gfs2_quota_change *qc,
-			  struct gfs2_buffer_head *bh)
+void gfs2_quota_change_in(struct gfs2_quota_change *qc, char *buf)
 {
-	struct gfs2_quota_change *str = (struct gfs2_quota_change *)(bh->b_data +
-		sizeof(struct gfs2_meta_header));
+	struct gfs2_quota_change *str = (struct gfs2_quota_change *)(buf +
+	                                 sizeof(struct gfs2_meta_header));
 
 	CPIN_64(qc, str, qc_change);
 	CPIN_32(qc, str, qc_flags);
 	CPIN_32(qc, str, qc_id);
 }
 
-void gfs2_quota_change_out(struct gfs2_quota_change *qc,
-			   struct gfs2_buffer_head *bh)
+void gfs2_quota_change_out(struct gfs2_quota_change *qc, char *buf)
 {
-	struct gfs2_quota_change *str = (struct gfs2_quota_change *)(bh->b_data +
-		sizeof(struct gfs2_meta_header));
+	struct gfs2_quota_change *str = (struct gfs2_quota_change *)(buf +
+	                                 sizeof(struct gfs2_meta_header));
 
 	CPOUT_64(qc, str, qc_change);
 	CPOUT_32(qc, str, qc_flags);
 	CPOUT_32(qc, str, qc_id);
-	bmodified(bh);
 }
 
 void gfs2_quota_change_print(const struct gfs2_quota_change *qc)
@@ -664,4 +713,3 @@ void gfs2_quota_change_print(const struct gfs2_quota_change *qc)
 	pv(qc, qc_id, "%u", "0x%x");
 }
 
-
