@@ -11,6 +11,7 @@
 #include "rgrp.h"
 
 #define RG_SYNC_TOLERANCE 1000
+#define ROUND_UP(N, S) ((((N) + (S) - 1) / (S)) * (S))
 
 static void compute_bitmaps(lgfs2_rgrp_t rg, const unsigned bsize)
 {
@@ -109,23 +110,30 @@ struct rgrp_tree *gfs2_blk2rgrpd(struct gfs2_sbd *sdp, uint64_t blk)
 int lgfs2_rgrp_bitbuf_alloc(lgfs2_rgrp_t rg)
 {
 	struct gfs2_sbd *sdp = rg->rgrps->sdp;
+	struct gfs2_buffer_head *bhs;
+	size_t len = rg->ri.ri_length * sdp->bsize;
+	unsigned long io_align = sdp->bsize;
 	unsigned i;
 	char *bufs;
 
-	bufs = calloc(rg->ri.ri_length, sizeof(struct gfs2_buffer_head) + sdp->bsize);
-	if (bufs == NULL)
+	if (rg->rgrps->align > 0) {
+		len = ROUND_UP(len, rg->rgrps->align * sdp->bsize);
+		io_align = rg->rgrps->align_off * sdp->bsize;
+	}
+	bhs = calloc(rg->ri.ri_length, sizeof(struct gfs2_buffer_head));
+	if (bhs == NULL)
 		return 1;
 
-	rg->bits[0].bi_bh = (struct gfs2_buffer_head *)bufs;
-	rg->bits[0].bi_bh->iov.iov_base = (char *)(rg->bits[0].bi_bh + 1);
-	rg->bits[0].bi_bh->iov.iov_len = sdp->bsize;
-	rg->bits[0].bi_bh->b_blocknr = rg->ri.ri_addr;
-	rg->bits[0].bi_bh->sdp = sdp;
+	if (posix_memalign((void **)&bufs, io_align, len) != 0) {
+		errno = ENOMEM;
+		free(bhs);
+		return 1;
+	}
+	memset(bufs, 0, len);
 
-	for (i = 1; i < rg->ri.ri_length; i++) {
-		char *nextbuf = rg->bits[i - 1].bi_bh->b_data + sdp->bsize;
-		rg->bits[i].bi_bh = (struct gfs2_buffer_head *)(nextbuf);
-		rg->bits[i].bi_bh->iov.iov_base = (char *)(rg->bits[i].bi_bh + 1);
+	for (i = 0; i < rg->ri.ri_length; i++) {
+		rg->bits[i].bi_bh = bhs + i;
+		rg->bits[i].bi_bh->iov.iov_base = bufs + (i * sdp->bsize);
 		rg->bits[i].bi_bh->iov.iov_len = sdp->bsize;
 		rg->bits[i].bi_bh->b_blocknr = rg->ri.ri_addr + i;
 		rg->bits[i].bi_bh->sdp = sdp;
@@ -143,11 +151,49 @@ int lgfs2_rgrp_bitbuf_alloc(lgfs2_rgrp_t rg)
 void lgfs2_rgrp_bitbuf_free(lgfs2_rgrp_t rg)
 {
 	unsigned i;
+	free(rg->bits[0].bi_bh->iov.iov_base);
 	free(rg->bits[0].bi_bh);
 	for (i = 0; i < rg->ri.ri_length; i++)
 		rg->bits[i].bi_bh = NULL;
 }
 
+/**
+ * Check a resource group's crc
+ * Returns 0 on success, non-zero if crc is bad
+ */
+int lgfs2_rgrp_crc_check(char *buf)
+{
+	int ret = 0;
+#ifdef GFS2_HAS_RG_RI_FIELDS
+	struct gfs2_rgrp *rg = (struct gfs2_rgrp *)buf;
+	uint32_t crc = rg->rg_crc;
+
+	if (crc == 0)
+		return 0;
+
+	rg->rg_crc = 0;
+	if (be32_to_cpu(crc) != gfs2_disk_hash(buf, sizeof(struct gfs2_rgrp)))
+		ret = 1;
+	rg->rg_crc = crc;
+#endif
+	return ret;
+}
+
+/**
+ * Set the crc of an on-disk resource group
+ */
+void lgfs2_rgrp_crc_set(char *buf)
+{
+#ifdef GFS2_HAS_RG_RI_FIELDS
+	struct gfs2_rgrp *rg = (struct gfs2_rgrp *)buf;
+	uint32_t crc;
+
+	rg->rg_crc = 0;
+	crc = gfs2_disk_hash(buf, sizeof(struct gfs2_rgrp));
+	rg->rg_crc = cpu_to_be32(crc);
+#endif
+}
+
 /**
  * gfs2_rgrp_read - read in the resource group information from disk.
  * @rgd - resource group structure
@@ -185,11 +231,14 @@ uint64_t gfs2_rgrp_read(struct gfs2_sbd *sdp, struct rgrp_tree *rgd)
 			return rgd->ri.ri_addr + err;
 		}
 	}
-	if (x > 0) {
-		if (sdp->gfs1)
-			gfs_rgrp_in((struct gfs_rgrp *)&rgd->rg, rgd->bits[0].bi_bh);
-		else
-			gfs2_rgrp_in(&rgd->rg, rgd->bits[0].bi_bh);
+	if (sdp->gfs1)
+		gfs_rgrp_in((struct gfs_rgrp *)&rgd->rg, rgd->bits[0].bi_bh);
+	else {
+		if (lgfs2_rgrp_crc_check(rgd->bits[0].bi_bh->b_data)) {
+			free(bhs);
+			return rgd->ri.ri_addr;
+		}
+		gfs2_rgrp_in(&rgd->rg, rgd->bits[0].bi_bh->b_data);
 	}
 	free(bhs);
 	return 0;
@@ -335,7 +384,7 @@ uint32_t lgfs2_rgrps_plan(const lgfs2_rgrps_t rgs, uint64_t space, uint32_t tgts
 		   over all of the rgrps so that we don't end with a single
 		   tiny one.  */
 		rgs->plan[0].num++;
-		while (((rgs->plan[0].len - adj) * (rgs->plan[0].num)) >= space)
+		while (((rgs->plan[0].len - adj) * (uint64_t)rgs->plan[0].num) >= space)
 			rgs->plan[0].len -= adj;
 
 		/* We've adjusted the size of the rgrps down as far as we can
@@ -421,7 +470,7 @@ unsigned lgfs2_rindex_read_fd(int fd, lgfs2_rgrps_t rgs)
 			return 0;
 
 		gfs2_rindex_in(&ri, buf);
-		rg = lgfs2_rgrps_append(rgs, &ri);
+		rg = lgfs2_rgrps_append(rgs, &ri, 0);
 		if (rg == NULL)
 			return 0;
 		count++;
@@ -454,7 +503,7 @@ const struct gfs2_rindex *lgfs2_rindex_read_one(struct gfs2_inode *rip, lgfs2_rg
 		return NULL;
 
 	gfs2_rindex_in(&ri, buf);
-	rg = lgfs2_rgrps_append(rgs, &ri);
+	rg = lgfs2_rgrps_append(rgs, &ri, 0);
 	if (rg == NULL)
 		return NULL;
 
@@ -573,9 +622,10 @@ struct osi_node *lgfs2_rgrps_root(lgfs2_rgrps_t rgs)
  * Insert a new resource group after the last resource group in a set.
  * rgs: The set of resource groups
  * entry: The entry to be added
+ * rg_skip: The value to be used for this resource group's rg_skip field
  * Returns the new resource group on success or NULL on failure with errno set.
  */
-lgfs2_rgrp_t lgfs2_rgrps_append(lgfs2_rgrps_t rgs, struct gfs2_rindex *entry)
+lgfs2_rgrp_t lgfs2_rgrps_append(lgfs2_rgrps_t rgs, struct gfs2_rindex *entry, uint32_t rg_skip)
 {
 	lgfs2_rgrp_t rg;
 	struct osi_node **link = &rgs->root.osi_node;
@@ -606,7 +656,15 @@ lgfs2_rgrp_t lgfs2_rgrps_append(lgfs2_rgrps_t rgs, struct gfs2_rindex *entry)
 	rg->rg.rg_header.mh_type = GFS2_METATYPE_RG;
 	rg->rg.rg_header.mh_format = GFS2_FORMAT_RG;
 	rg->rg.rg_free = rg->ri.ri_data;
-
+#ifdef GFS2_HAS_RG_SKIP
+	rg->rg.rg_skip = rg_skip;
+#endif
+#ifdef GFS2_HAS_RG_RI_FIELDS
+	rg->rg.rg_data0 = rg->ri.ri_data0;
+	rg->rg.rg_data = rg->ri.ri_data;
+	rg->rg.rg_bitbytes = rg->ri.ri_bitbytes;
+	rg->rg.rg_crc = 0;
+#endif
 	compute_bitmaps(rg, rgs->sdp->bsize);
 	rg->rgrps = rgs;
 	return rg;
@@ -618,7 +676,7 @@ lgfs2_rgrp_t lgfs2_rgrps_append(lgfs2_rgrps_t rgs, struct gfs2_rindex *entry)
  */
 int lgfs2_rgrp_write(int fd, const lgfs2_rgrp_t rg)
 {
-	int ret = 0;
+	struct gfs2_sbd *sdp = rg->rgrps->sdp;
 	unsigned int i;
 	const struct gfs2_meta_header bmh = {
 		.mh_magic = GFS2_MAGIC,
@@ -626,25 +684,29 @@ int lgfs2_rgrp_write(int fd, const lgfs2_rgrp_t rg)
 		.mh_format = GFS2_FORMAT_RB,
 	};
 	int freebufs = 0;
+	ssize_t ret;
+	size_t len;
 
 	if (rg->bits[0].bi_bh == NULL) {
 		freebufs = 1;
 		if (lgfs2_rgrp_bitbuf_alloc(rg) != 0)
 			return -1;
 	}
-
 	gfs2_rgrp_out(&rg->rg, rg->bits[0].bi_bh->b_data);
-	ret = bwrite(rg->bits[0].bi_bh);
-
-	for (i = 1; ret == 0 && i < rg->ri.ri_length; i++) {
+	for (i = 1; i < rg->ri.ri_length; i++)
 		gfs2_meta_header_out(&bmh, rg->bits[i].bi_bh->b_data);
-		ret = bwrite(rg->bits[i].bi_bh);
-	}
+
+	len = sdp->bsize * rg->ri.ri_length;
+	if (rg->rgrps->align > 0)
+		len = ROUND_UP(len, rg->rgrps->align * sdp->bsize);
+
+	ret = pwrite(sdp->device_fd, rg->bits[0].bi_bh->b_data, len,
+	             rg->bits[0].bi_bh->b_blocknr * sdp->bsize);
 
 	if (freebufs)
 		lgfs2_rgrp_bitbuf_free(rg);
 
-	return ret;
+	return ret == len ? 0 : -1;
 }
 
 lgfs2_rgrp_t lgfs2_rgrp_first(lgfs2_rgrps_t rgs)
@@ -657,6 +719,11 @@ lgfs2_rgrp_t lgfs2_rgrp_next(lgfs2_rgrp_t rg)
 	return (lgfs2_rgrp_t)osi_next(&rg->node);
 }
 
+lgfs2_rgrp_t lgfs2_rgrp_prev(lgfs2_rgrp_t rg)
+{
+	return (lgfs2_rgrp_t)osi_prev(&rg->node);
+}
+
 lgfs2_rgrp_t lgfs2_rgrp_last(lgfs2_rgrps_t rgs)
 {
 	return (lgfs2_rgrp_t)osi_last(&rgs->root);
