@@ -30,6 +30,10 @@
 #include "gfs2_mkfs.h"
 #include "progress.h"
 
+#ifdef GFS2_HAS_UUID
+#include <uuid.h>
+#endif
+
 static void print_usage(const char *prog_name)
 {
 	int i;
@@ -132,6 +136,7 @@ struct mkfs_opts {
 	unsigned got_lockproto:1;
 	unsigned got_locktable:1;
 	unsigned got_device:1;
+	unsigned got_topol:1;
 
 	unsigned override:1;
 	unsigned quiet:1;
@@ -252,6 +257,33 @@ static unsigned parse_bool(struct mkfs_opts *opts, const char *key, const char *
 	exit(-1);
 }
 
+static int parse_topology(struct mkfs_opts *opts, char *str)
+{
+	char *opt;
+	unsigned i = 0;
+	unsigned long *topol[5] = {
+		&opts->dev.alignment_offset,
+		&opts->dev.logical_sector_size,
+		&opts->dev.minimum_io_size,
+		&opts->dev.optimal_io_size,
+		&opts->dev.physical_sector_size
+	};
+
+	while ((opt = strsep(&str, ":")) != NULL) {
+		if (i > 4) {
+			fprintf(stderr, "Too many topology values.\n");
+			return 1;
+		}
+		*topol[i] = parse_ulong(opts, "test_topology", opt);
+		i++;
+	}
+	if (i < 5) {
+		fprintf(stderr, "Too few topology values.\n");
+		return 1;
+	}
+	return 0;
+}
+
 static void opt_parse_extended(char *str, struct mkfs_opts *opts)
 {
 	char *opt;
@@ -270,6 +302,11 @@ static void opt_parse_extended(char *str, struct mkfs_opts *opts)
 			opts->got_swidth = 1;
 		} else if (strcmp("align", key) == 0) {
 			opts->align = parse_bool(opts, "align", val);
+		} else if (strcmp("test_topology", key) == 0) {
+			if (parse_topology(opts, val) != 0)
+				exit(-1);
+			opts->got_topol = (opts->dev.logical_sector_size != 0 &&
+	                                   opts->dev.physical_sector_size != 0);
 		} else if (strcmp("help", key) == 0) {
 			print_ext_opts();
 			exit(0);
@@ -422,34 +459,29 @@ static void test_locking(struct mkfs_opts *opts)
 
 static void are_you_sure(void)
 {
-	char *line = NULL;
-	size_t len = 0;
-	int ret = -1;
-	int res = 0;
+	while (1) {
+		char *line = NULL;
+		size_t len = 0;
+		int ret;
+		int res;
 
-	do{
 		/* Translators: We use rpmatch(3) to match the answers to y/n
 		   questions in the user's own language, so the [y/n] here must also be
 		   translated to match one of the letters in the pattern printed by
 		   `locale -k yesexpr` and one of the letters in the pattern printed by
 		   `locale -k noexpr` */
-		printf( _("Are you sure you want to proceed? [y/n]"));
+		printf( _("Are you sure you want to proceed? [y/n] "));
 		ret = getline(&line, &len, stdin);
 		res = rpmatch(line);
-		
-		if (res > 0){
-			free(line);
-			return;
-		}
-		if (!res){
-			printf("\n");
-			die( _("Aborted.\n"));
-		}
-		
-	}while(ret >= 0);
-
-	if(line)
 		free(line);
+		if (ret <= 0)
+			continue;
+		if (res == 1) /* Yes */
+			return;
+		if (res == 0) /* No */
+			exit(1);
+		/* Unrecognized input; go again. */
+	};
 }
 
 static unsigned choose_blocksize(struct mkfs_opts *opts)
@@ -457,16 +489,21 @@ static unsigned choose_blocksize(struct mkfs_opts *opts)
 	unsigned int x;
 	unsigned int bsize = opts->bsize;
 	struct mkfs_dev *dev = &opts->dev;
+	int got_topol = (dev->got_topol || opts->got_topol);
 
-	if (dev->got_topol && opts->debug) {
+	if (got_topol && opts->debug) {
 		printf("alignment_offset: %lu\n", dev->alignment_offset);
 		printf("logical_sector_size: %lu\n", dev->logical_sector_size);
 		printf("minimum_io_size: %lu\n", dev->minimum_io_size);
 		printf("optimal_io_size: %lu\n", dev->optimal_io_size);
 		printf("physical_sector_size: %lu\n", dev->physical_sector_size);
 	}
-
-	if (!opts->got_bsize && dev->got_topol) {
+	if (got_topol && dev->alignment_offset != 0) {
+		fprintf(stderr,
+		  _("Warning: device is not properly aligned. This may harm performance.\n"));
+		dev->physical_sector_size = dev->logical_sector_size;
+	}
+	if (!opts->got_bsize && got_topol) {
 		if (dev->optimal_io_size <= getpagesize() &&
 		    dev->optimal_io_size >= dev->minimum_io_size)
 			bsize = dev->optimal_io_size;
@@ -515,7 +552,7 @@ static void opts_check(struct mkfs_opts *opts)
 	if (!opts->journals)
 		die( _("no journals specified\n"));
 
-	if (opts->jsize < 8 || opts->jsize > 1024)
+	if (opts->jsize < GFS2_MIN_JSIZE || opts->jsize > GFS2_MAX_JSIZE)
 		die( _("bad journal size\n"));
 
 	if (!opts->qcsize || opts->qcsize > 64)
@@ -538,12 +575,18 @@ static void print_results(struct gfs2_sb *sb, struct mkfs_opts *opts, uint64_t r
 	printf("%-27s%.2f %s (%"PRIu64" %s)\n", _("Filesystem size:"),
 	       (fssize / ((float)(1 << 30)) * sb->sb_bsize), _("GB"), fssize, _("blocks"));
 	printf("%-27s%u\n", _("Journals:"), opts->journals);
+	printf("%-27s%uMB\n", _("Journal size:"), opts->jsize);
 	printf("%-27s%"PRIu64"\n", _("Resource groups:"), rgrps);
 	printf("%-27s\"%s\"\n", _("Locking protocol:"), opts->lockproto);
 	printf("%-27s\"%s\"\n", _("Lock table:"), opts->locktable);
 #ifdef GFS2_HAS_UUID
+	{
+	char readable_uuid[36+1];
+
+	uuid_unparse(sb->sb_uuid, readable_uuid);
 	/* Translators: "UUID" = universally unique identifier. */
-	printf("%-27s%s\n", _("UUID:"), str_uuid(sb->sb_uuid));
+	printf("%-27s%s\n", _("UUID:"), readable_uuid);
+	}
 #endif
 }
 
@@ -570,6 +613,42 @@ static void warn_of_destruction(const char *path)
 	free(abspath);
 }
 
+static int zero_gap(struct gfs2_sbd *sdp, uint64_t addr, size_t blocks)
+{
+	struct iovec *iov;
+	char *zerobuf;
+	ssize_t wrote;
+	unsigned i;
+
+	if (blocks == 0)
+		return 0;
+	iov = calloc(blocks, sizeof(*iov));
+	if (iov == NULL) {
+		perror(_("Failed to zero blocks\n"));
+		return 1;
+	}
+	zerobuf = calloc(1, sdp->bsize);
+	if (zerobuf == NULL) {
+		perror(_("Failed to zero blocks\n"));
+		free(iov);
+		return 1;
+	}
+	for (i = 0; i < blocks; i++) {
+		iov[i].iov_base = zerobuf;
+		iov[i].iov_len = sdp->bsize;
+	}
+	wrote = pwritev(sdp->device_fd, iov, blocks, addr * sdp->bsize);
+	if (wrote != blocks * sdp->bsize) {
+		fprintf(stderr, _("Zeroing write failed at block %"PRIu64"\n"), addr);
+		free(zerobuf);
+		free(iov);
+		return 1;
+	}
+	free(zerobuf);
+	free(iov);
+	return 0;
+}
+
 static lgfs2_rgrps_t rgs_init(struct mkfs_opts *opts, struct gfs2_sbd *sdp)
 {
 	lgfs2_rgrps_t rgs;
@@ -590,8 +669,14 @@ static lgfs2_rgrps_t rgs_init(struct mkfs_opts *opts, struct gfs2_sbd *sdp)
 			al_off = opts->sunit / sdp->bsize;
 		}
 	} else if (opts->align) {
-		if ((opts->dev.minimum_io_size > opts->dev.physical_sector_size) &&
-		    (opts->dev.optimal_io_size > opts->dev.physical_sector_size)) {
+		if (opts->dev.optimal_io_size <= opts->dev.physical_sector_size ||
+		    opts->dev.minimum_io_size <= opts->dev.physical_sector_size ||
+		    (opts->dev.optimal_io_size % opts->dev.minimum_io_size) != 0) {
+			/* If optimal_io_size is not a multiple of minimum_io_size then
+			   the values are not reliable swidth and sunit values, so disable
+			   rgrp alignment */
+			opts->align = 0;
+		} else {
 			al_base = opts->dev.optimal_io_size / sdp->bsize;
 			al_off = opts->dev.minimum_io_size / sdp->bsize;
 		}
@@ -616,9 +701,18 @@ static lgfs2_rgrps_t rgs_init(struct mkfs_opts *opts, struct gfs2_sbd *sdp)
 
 static int place_rgrp(struct gfs2_sbd *sdp, lgfs2_rgrp_t rg, int debug)
 {
-	int err = 0;
+	uint64_t prev_end = (GFS2_SB_ADDR * GFS2_BASIC_BLOCK / sdp->bsize) + 1;
 	const struct gfs2_rindex *ri = lgfs2_rgrp_index(rg);
+	lgfs2_rgrp_t prev = lgfs2_rgrp_prev(rg);
+	int err = 0;
 
+	if (prev != NULL) {
+		prev_end = lgfs2_rgrp_index(prev)->ri_data0 +
+		           lgfs2_rgrp_index(prev)->ri_data;
+	}
+	err = zero_gap(sdp, prev_end, ri->ri_addr - prev_end);
+	if (err != 0)
+		return -1;
 	err = lgfs2_rgrp_write(sdp->device_fd, rg);
 	if (err != 0) {
 		perror(_("Failed to write resource group"));
@@ -637,11 +731,12 @@ static int place_rgrp(struct gfs2_sbd *sdp, lgfs2_rgrp_t rg, int debug)
 static int add_rgrp(lgfs2_rgrps_t rgs, uint64_t *addr, uint32_t len, lgfs2_rgrp_t *rg)
 {
 	struct gfs2_rindex ri;
+	uint64_t nextaddr;
 
 	/* When we get to the end of the device, it's only an error if we have
 	   more structures left to write, i.e. when len is != 0. */
-	*addr = lgfs2_rindex_entry_new(rgs, &ri, *addr, len);
-	if (*addr == 0) {
+	nextaddr = lgfs2_rindex_entry_new(rgs, &ri, *addr, len);
+	if (nextaddr == 0) {
 		if (len != 0) {
 			perror(_("Failed to create resource group index entry"));
 			return -1;
@@ -649,12 +744,12 @@ static int add_rgrp(lgfs2_rgrps_t rgs, uint64_t *addr, uint32_t len, lgfs2_rgrp_
 			return 1;
 		}
 	}
-
-	*rg = lgfs2_rgrps_append(rgs, &ri);
+	*rg = lgfs2_rgrps_append(rgs, &ri, nextaddr - *addr);
 	if (*rg == NULL) {
 		perror(_("Failed to create resource group"));
 		return -1;
 	}
+	*addr = nextaddr;
 	return 0;
 }
 
@@ -665,21 +760,20 @@ static int place_journals(struct gfs2_sbd *sdp, lgfs2_rgrps_t rgs, struct mkfs_o
 	uint32_t rgsize = lgfs2_rgsize_for_data(jfsize, sdp->bsize);
 	unsigned j;
 
-	/* Initialise a progress bar for resource group creation. */
 	gfs2_progress_init(&progress, opts->journals, _("Adding journals: "), opts->quiet);
 
 	/* We'll build the jindex later so remember where we put the journals */
 	mkfs_journals = calloc(opts->journals, sizeof(*mkfs_journals));
 	if (mkfs_journals == NULL)
 		return 1;
-	*rgaddr = lgfs2_rgrp_align_addr(rgs, sdp->sb_addr + 1);
+	*rgaddr = lgfs2_rgrp_align_addr(rgs, LGFS2_SB_ADDR(sdp) + 1);
+	rgsize = lgfs2_rgrp_align_len(rgs, rgsize);
 
 	for (j = 0; j < opts->journals; j++) {
 		int result;
 		lgfs2_rgrp_t rg;
 		struct gfs2_inode in = {0};
 
-		/* Update progress bar for journal creation. */
 		gfs2_progress_update(&progress, (j + 1));
 
 		if (opts->debug)
@@ -734,26 +828,19 @@ static int place_journals(struct gfs2_sbd *sdp, lgfs2_rgrps_t rgs, struct mkfs_o
 	return 0;
 }
 
-static int place_rgrps(struct gfs2_sbd *sdp, lgfs2_rgrps_t rgs, struct mkfs_opts *opts)
+static int place_rgrps(struct gfs2_sbd *sdp, lgfs2_rgrps_t rgs, uint64_t *rgaddr, struct mkfs_opts *opts)
 {
 	struct gfs2_progress_bar progress;
-	uint64_t rgaddr = lgfs2_rgrp_align_addr(rgs, sdp->sb_addr + 1);
 	uint32_t rgblks = ((opts->rgsize << 20) / sdp->bsize);
 	uint32_t rgnum;
 	int result;
 
-	result = place_journals(sdp, rgs, opts, &rgaddr);
-	if (result != 0)
-		return result;
-
-	rgnum = lgfs2_rgrps_plan(rgs, sdp->device.length - rgaddr, rgblks);
-
-	/* Initialise a progress bar for resource group creation (after journal creation). */
+	rgnum = lgfs2_rgrps_plan(rgs, sdp->device.length - *rgaddr, rgblks);
 	gfs2_progress_init(&progress, (rgnum + opts->journals), _("Building resource groups: "), opts->quiet);
 
 	while (1) {
 		lgfs2_rgrp_t rg;
-		result = add_rgrp(rgs, &rgaddr, 0, &rg);
+		result = add_rgrp(rgs, rgaddr, 0, &rg);
 		if (result > 0)
 			break;
 		else if (result < 0)
@@ -765,14 +852,44 @@ static int place_rgrps(struct gfs2_sbd *sdp, lgfs2_rgrps_t rgs, struct mkfs_opts
 			return result;
 		}
 
-		/* Update progress bar with resource group address. */
 		gfs2_progress_update(&progress, (sdp->rgrps));
 	}
 	gfs2_progress_close(&progress, _("Done\n"));
+	posix_fadvise(sdp->device_fd, 0, sdp->fssize * sdp->bsize, POSIX_FADV_DONTNEED);
 
 	return 0;
 }
 
+/*
+ * Find a reasonable journal file size (in blocks) given the number of blocks
+ * in the filesystem.  For very small filesystems, it is not reasonable to
+ * have a journal that fills more than half of the filesystem.
+ *
+ * n.b. comments assume 4k blocks
+ *
+ * This was copied and adapted from e2fsprogs.
+ */
+static int default_journal_size(unsigned bsize, uint64_t num_blocks)
+{
+	int min_blocks = (GFS2_MIN_JSIZE << 20) / bsize;
+
+	if (num_blocks < 2 * min_blocks)
+		return -1;
+	if (num_blocks < 131072)        /* 512 MB */
+		return min_blocks;              /* 8 MB */
+	if (num_blocks < 512*1024)      /* 2 GB */
+		return (4096);                  /* 16 MB */
+	if (num_blocks < 2048*1024)     /* 8 GB */
+		return (8192);                  /* 32 MB */
+	if (num_blocks < 4096*1024)     /* 16 GB */
+		return (16384);                 /* 64 MB */
+	if (num_blocks < 262144*1024)   /*  1 TB */
+		return (32768);                 /* 128 MB */
+	if (num_blocks < 2621440UL*1024)  /* 10 TB */
+		return (131072);                /* 512 MB */
+	return 262144;                          /*   1 GB */
+}
+
 static void sbd_init(struct gfs2_sbd *sdp, struct mkfs_opts *opts, unsigned bsize)
 {
 	memset(sdp, 0, sizeof(struct gfs2_sbd));
@@ -797,9 +914,28 @@ static void sbd_init(struct gfs2_sbd *sdp, struct mkfs_opts *opts, unsigned bsiz
 			       opts->dev.size / ((float)(1 << 30)), _("GB"),
 			       opts->dev.size / sdp->bsize, _("blocks"));
 		}
-		/* TODO: Check if the fssize is too small, somehow */
 		sdp->device.length = opts->fssize;
 	}
+	/* opts->jsize has already been max/min checked but we need to check it
+	   makes sense for the device size, or set a sensible default, if one
+	   will fit. For user-provided journal sizes, limit it to half of the fs. */
+	if (!opts->got_jsize) {
+		int default_jsize = default_journal_size(sdp->bsize, sdp->device.length / opts->journals);
+		if (default_jsize < 0) {
+			fprintf(stderr, _("gfs2 will not fit on this device.\n"));
+			exit(1);
+		}
+		opts->jsize = (default_jsize * sdp->bsize) >> 20;
+	} else if ((((opts->jsize * opts->journals) << 20) / sdp->bsize) > (sdp->device.length / 2)) {
+		unsigned max_jsize = (sdp->device.length / 2 * sdp->bsize / opts->journals) >> 20;
+
+		fprintf(stderr, _("gfs2 will not fit on this device.\n"));
+		if (max_jsize >= GFS2_MIN_JSIZE)
+			fprintf(stderr, _("Maximum size for %u journals on this device is %uMB.\n"),
+			        opts->journals, max_jsize);
+		exit(1);
+	}
+	sdp->jsize = opts->jsize;
 }
 
 static int probe_contents(struct mkfs_dev *dev)
@@ -842,7 +978,8 @@ static int probe_contents(struct mkfs_dev *dev)
 			dev->minimum_io_size = blkid_topology_get_minimum_io_size(tp);
 			dev->optimal_io_size = blkid_topology_get_optimal_io_size(tp);
 			dev->physical_sector_size = blkid_topology_get_physical_sector_size(tp);
-			dev->got_topol = 1;
+			dev->got_topol = (dev->logical_sector_size != 0 &&
+	                                  dev->physical_sector_size != 0);
 		}
 	}
 
@@ -850,7 +987,7 @@ static int probe_contents(struct mkfs_dev *dev)
 	return 0;
 }
 
-static void open_dev(struct mkfs_dev *dev)
+static void open_dev(struct mkfs_dev *dev, int withprobe)
 {
 	int error;
 
@@ -860,6 +997,11 @@ static void open_dev(struct mkfs_dev *dev)
 		exit(1);
 	}
 
+	/* Freshen up the cache */
+	posix_fadvise(dev->fd, 0, 0, POSIX_FADV_DONTNEED);
+	/* Turn off readahead, we're just writing new blocks */
+	posix_fadvise(dev->fd, 0, 0, POSIX_FADV_RANDOM);
+
 	error = fstat(dev->fd, &dev->stat);
 	if (error < 0) {
 		perror(dev->path);
@@ -878,9 +1020,7 @@ static void open_dev(struct mkfs_dev *dev)
 		fprintf(stderr, _("'%s' is not a block device or regular file\n"), dev->path);
 		exit(1);
 	}
-
-	error = probe_contents(dev);
-	if (error)
+	if (withprobe && (probe_contents(dev) != 0))
 		exit(1);
 }
 
@@ -890,6 +1030,7 @@ int main(int argc, char *argv[])
 	struct gfs2_sb sb;
 	struct mkfs_opts opts;
 	lgfs2_rgrps_t rgs;
+	uint64_t rgaddr;
 	int error;
 	unsigned bsize;
 
@@ -901,7 +1042,7 @@ int main(int argc, char *argv[])
 	opts_get(argc, argv, &opts);
 	opts_check(&opts);
 
-	open_dev(&opts.dev);
+	open_dev(&opts.dev, !opts.got_topol);
 	bsize = choose_blocksize(&opts);
 
 	if (S_ISREG(opts.dev.stat.st_mode)) {
@@ -939,8 +1080,13 @@ int main(int argc, char *argv[])
 		if (!opts.quiet)
 			printf("%s", _("Done\n"));
 	}
-
-	error = place_rgrps(&sbd, rgs, &opts);
+	rgaddr = lgfs2_rgrp_align_addr(rgs, LGFS2_SB_ADDR(&sbd) + 1);
+	error = place_journals(&sbd, rgs, &opts, &rgaddr);
+	if (error != 0) {
+		fprintf(stderr, _("Failed to create journals\n"));
+		exit(1);
+	}
+	error = place_rgrps(&sbd, rgs, &rgaddr, &opts);
 	if (error) {
 		fprintf(stderr, _("Failed to build resource groups\n"));
 		exit(1);
@@ -1028,7 +1174,7 @@ int main(int argc, char *argv[])
 		perror(opts.dev.path);
 		exit(EXIT_FAILURE);
 	}
-
+	posix_fadvise(opts.dev.fd, 0, 0, POSIX_FADV_DONTNEED);
 	error = close(opts.dev.fd);
 	if (error){
 		perror(opts.dev.path);
